@@ -1,16 +1,15 @@
 import json
+from pathlib import Path
 import re
-import logging
 from functools import wraps
 from shutil import which
-from typing import Optional, Union, List, Dict, Callable
+import tempfile
+from typing import Optional, Tuple, Union, List, Dict, Callable
 
 from rclone_python import utils
 from rclone_python.hash_types import HashTypes
 from rclone_python.remote_types import RemoteTypes
-
-# debug flag enables/disables raw output of rclone progresses in the terminal
-DEBUG = False
+from rclone_python.logs import logger
 
 
 def __check_installed(func):
@@ -26,32 +25,69 @@ def __check_installed(func):
     return wrapper
 
 
+def set_config_file(config_file: str):
+    """Change the config file used by this wrapper."""
+    utils.Config(config_file)
+
+
+def set_executable_file(executable_file: str, validate: bool = True):
+    """Change the rclone executable path used by this wrapper.
+
+    Args:
+        executable_file (str): The path to the rclone executable.
+        validate (bool): If True, validates that the path exists. Defaults to True.
+
+    Raises:
+        FileNotFoundError: If validate is True and the executable path does not exist.
+    """
+    if validate:
+        path = Path(executable_file)
+        if not path.is_file():
+            raise FileNotFoundError(
+                f"Executable path '{executable_file}' does not exist"
+            )
+
+    config = utils.Config()
+    config.executable_path = executable_file
+
+
+def set_log_level(level: int):
+    """Change the log level of this wrapper.
+
+    Args:
+        level (int): The log level to use.
+    """
+    logger.setLevel(level)
+
+
 def is_installed() -> bool:
     """
-    :return: True if rclone is correctly installed on the system.
+    :return: True if rclone is correctly installed on the system or a valid custom executable path is set.
     """
-    return which("rclone") is not None
+    # Check if rclone is in system PATH
+    if which("rclone") is not None:
+        return True
+
+    # Check if a custom executable path was set
+    config = utils.Config()
+    if config.executable_path != "rclone":
+        return Path(config.executable_path).is_file()
+
+    return False
 
 
 @__check_installed
-def about(remote_name: str):
+def about(path: str) -> Dict:
     """
     Executes the rclone about command and returns the retrieved json as a dictionary.
+    All sizes are in number of bytes. This command can be run with local and remote paths.
     :param remote_name: The name of the remote to examine.
     :return: Dictionary with remote properties.
     """
-    if not remote_name.endswith(":"):
-        # if the remote name missed the colon manually add it.
-        remote_name += ":"
 
-    process = utils.run_cmd(f"rclone about {remote_name} --json")
+    stdout, _ = utils.run_rclone_cmd(f'about "{path}" --json')
 
-    if process.returncode == 0:
-        return json.loads(process.stdout)
-    else:
-        raise Exception(
-            f"An error occurred while executing the about command: {process.stderr}"
-        )
+    return json.loads(stdout)
 
 
 @__check_installed
@@ -93,15 +129,15 @@ def create_remote(
 
     if not check_remote_existing(remote_name):
         # set up the selected cloud
-        command = f'rclone config create "{remote_name}" "{remote_type}"'
+        command = f'config create "{remote_name}" "{remote_type}"'
 
         if client_id and client_secret:
-            logging.info("Using the provided client id and client secret.")
+            logger.info("Using the provided client id and client secret.")
 
             kwargs["client_id"] = client_id
             kwargs["client_secret"] = client_secret
         else:
-            logging.warning(
+            logger.warning(
                 "The drive client id and the client secret have not been set. Using defaults."
             )
 
@@ -110,16 +146,14 @@ def create_remote(
             command += f' {key}="{value}"'
 
         # run the setup command
-        process = utils.run_cmd(command)
-
-        if process.returncode != 0:
-            raise Exception(process.stderr)
+        utils.run_rclone_cmd(command)
     else:
         raise Exception(
             f"A rclone remote with the name '{remote_name}' already exists!"
         )
 
 
+@__check_installed
 def mkdir(
     path: str,
     args=None,
@@ -132,16 +166,10 @@ def mkdir(
     if args is None:
         args = []
 
-    process = utils.run_cmd(f"rclone mkdir {path}", args=args)
-
-    if process.returncode == 0:
-        return process.stdout
-    else:
-        raise Exception(
-            f"An error occurred while executing the mkdir command: {process.stderr}"
-        )
+    utils.run_rclone_cmd(f'mkdir "{path}"', args=args)
 
 
+@__check_installed
 def cat(
     path: str,
     count: Optional[int] = None,
@@ -171,14 +199,8 @@ def cat(
     if tail is not None:
         args.append(f"--tail {tail}")
 
-    process = utils.run_cmd(f"rclone cat {path}", args=args)
-
-    if process.returncode == 0:
-        return process.stdout
-    else:
-        raise Exception(
-            f"An error occurred while executing the cat command: {process.stderr}"
-        )
+    stdout, _ = utils.run_rclone_cmd(f'cat "{path}"', args=args)
+    return stdout
 
 
 def copy(
@@ -358,8 +380,9 @@ def get_remotes() -> List[str]:
     """
     :return: A list of all available remotes.
     """
-    command = "rclone listremotes"
-    remotes = utils.run_cmd(command).stdout.split()
+    command = "listremotes"
+    stdout, _ = utils.run_rclone_cmd(command)
+    remotes = stdout.split()
     if remotes is None:
         remotes = []
 
@@ -376,14 +399,8 @@ def purge(path: str, args=None):
     if args is None:
         args = []
 
-    command = f'rclone purge "{path}"'
-    process = utils.run_cmd(command, args)
-    if process.returncode == 0:
-        logging.info(f"Successfully purged {path}")
-    else:
-        raise Exception(
-            f'Purging path "{path}" failed with error message:\n{process.stderr}'
-        )
+    command = f'purge "{path}"'
+    utils.run_rclone_cmd(command, args)
 
 
 @__check_installed
@@ -397,15 +414,8 @@ def delete(path: str, args=None):
     if args is None:
         args = []
 
-    command = f'rclone delete "{path}"'
-    process = utils.run_cmd(command, args)
-
-    if process.returncode == 0:
-        logging.info(f"Successfully deleted {path}")
-    else:
-        raise Exception(
-            f'Deleting path "{path}" failed with error message:\n{process.stderr}'
-        )
+    command = f'delete "{path}"'
+    utils.run_rclone_cmd(command, args)
 
 
 @__check_installed
@@ -426,7 +436,7 @@ def link(
     if args is None:
         args = []
 
-    command = f'rclone link "{path}"'
+    command = f'link "{path}"'
 
     # add optional parameters
     if expire is not None:
@@ -434,12 +444,9 @@ def link(
     if unlink:
         args.append(f"--unlink")
 
-    process = utils.run_cmd(command, args)
+    stdout, _ = utils.run_rclone_cmd(command, args)
 
-    if process.returncode != 0:
-        raise Exception(process.stderr)
-    else:
-        return process.stdout
+    return stdout
 
 
 @__check_installed
@@ -463,7 +470,7 @@ def ls(
     if args is None:
         args = []
 
-    command = f'rclone lsjson "{path}"'
+    command = f'lsjson "{path}"'
 
     # add optional parameters
     if max_depth is not None:
@@ -473,14 +480,32 @@ def ls(
     if files_only:
         args.append("--files-only")
 
-    process = utils.run_cmd(command, args)
-
-    if process.returncode == 0:
-        return json.loads(process.stdout)
-    else:
-        raise Exception(f"ls operation on {path} failed with:\n{process.stderr}")
+    stdout, _ = utils.run_rclone_cmd(command, args)
+    return json.loads(stdout)
 
 
+@__check_installed
+def size(
+    path: str,
+    args: List[str] = None,
+) -> Dict:
+    """Returns the total size and number of objects in the path.
+
+    Args:
+        path (str): The path to calculate the total size on.
+        args (List[str], optional): Optional additional list of flags.
+
+    Returns:
+        Dict: Dictionary containing the file count, total file size in bytes and number of empty items.
+    """
+    if args is None:
+        args = []
+
+    stdout, _ = utils.run_rclone_cmd(f'size "{path}" --json', args)
+    return json.loads(stdout)
+
+
+@__check_installed
 def tree(
     path: str,
     args: List[str] = None,
@@ -497,12 +522,8 @@ def tree(
     if args is None:
         args = []
 
-    process = utils.run_cmd(f'rclone tree "{path}"', args)
-
-    if process.returncode != 0:
-        raise Exception(process.stderr)
-    else:
-        return process.stdout
+    stdout, _ = utils.run_rclone_cmd(f'tree "{path}"', args)
+    return stdout
 
 
 @__check_installed
@@ -525,7 +546,7 @@ def hash(
         args (List[str], optional): Optional additional list of flags.
 
     Raises:
-        Exception: Raised when the rclone command does not succeed.
+        RcloneException: Raised when the rclone command does not succeed.
 
     Returns:
         Union[None, str, bool, Dict[str, str], Dict[str, bool]]: 3 different modes apply based on the inputs:
@@ -553,13 +574,15 @@ def hash(
     if output_file is not None:
         args.append(f'--output-file "{output_file}"')
 
-    process: str = utils.run_cmd(f'rclone hashsum "{hash}" "{path}"', args)
+    returncode, stdout, stderr = utils.run_rclone_cmd(
+        f'hashsum "{hash}" "{path}"', args, raise_errors=False
+    )
 
-    lines = process.stdout.splitlines()
+    lines = stdout.splitlines()
 
     exception = False
 
-    if process.returncode != 0:
+    if returncode != 0:
         if checkfile is None:
             exception = True
         else:
@@ -570,8 +593,9 @@ def hash(
                     break
 
     if exception:
-        raise Exception(
-            f"hashsum operation on {path} with hash='{hash}' failed with:\n{process.stderr}"
+        raise utils.RcloneException(
+            f'hashsum command failed on path "{path}" with hash="{hash}"',
+            stderr,
         )
 
     if output_file is None:
@@ -580,7 +604,9 @@ def hash(
 
         for l in lines:
             if len(l) > 0:
-                value, key = l.split("  ", maxsplit=1)
+                # in checkfile mode only a single space separates key and value (key=matching, value=filename)
+                # while in normal mode a double space is used.
+                value, key = [item.strip() for item in l.split(" ", maxsplit=1)]
 
                 if checkfile is None:
                     hashsums[key] = value
@@ -595,11 +621,87 @@ def hash(
         return hashsums
 
 
+@__check_installed
+def check(
+    source: str,
+    dest: str,
+    combined: str = None,
+    size_only: bool = False,
+    download: bool = False,
+    one_way: bool = False,
+    args: List[str] = None,
+) -> Tuple[bool, List[Tuple[str, str]]]:
+    """Checks the files in the source and destination match.
+
+    Args:
+        source (str): The source path.
+        dest (str): The destination path.
+        combined (str, optional): Path to the combined file. Defaults to None.
+        size_only (bool, optional): Only compare the sizes not the hashes as well. Use this for a quick check. Defaults to False.
+        download (bool, optional): Download the data from both remotes and check them against each other on the fly. This can be useful for remotes that don't support hashes or if you really want to check all the data. Defaults to False.
+        one_way (bool, optional): Only check that files in the source match the files in the destination, not the other way around. This means that extra files in the destination that are not in the source will not be detected. Defaults to False.
+        args (List[str], optional): Optional additional list of flags and arguments. Defaults to None.
+
+    Raises:
+        utils.RcloneException: Raised when the rclone command does not succeed.
+
+    Returns:
+        Tuple[bool, List[Tuple[str, str]]]: The bool is true if source and dest match.
+                                            The list contains a symbol and all file paths in both directories. The following symbols are used:
+                                                "=" path means path was found in source and destination and was identical
+                                                "-" path means path was missing on the source, so only in the destination
+                                                "+" path means path was missing on the destination, so only in the source
+                                                "*" path means path was present in source and destination but different.
+                                                "!" path means there was an error reading or hashing the source or dest.
+
+    """
+    if args is None:
+        args = []
+    if size_only:
+        args.append("--size-only")
+    if download:
+        args.append("--download")
+    if one_way:
+        args.append("--one-way")
+
+    tmp = None
+    if not combined:
+        tmp = tempfile.TemporaryDirectory()
+        combined = Path(tmp.name, "combined_file")
+    # even if --combined is also specified by the user through args,
+    # this one will be used as apparently rclone uses the last specification.
+    args.append(f'--combined "{combined}"')
+
+    returncode, _, stderr = utils.run_rclone_cmd(
+        f'check "{source}" "{dest}"', args, raise_errors=False
+    )
+
+    logger.debug(f"Rclone check stderr output:\n{stderr}")
+
+    # read the combined file and extract all elements
+    combined_file = Path(combined)
+    if returncode != 0 and not combined_file.is_file():
+        raise utils.RcloneException(
+            f'check command failed on source: "{source}" dest: "{dest}"',
+            stderr,
+        )
+    out = [
+        # the file holds the symbol followed by a space and then the filepath
+        tuple(line.split(" ", maxsplit=1))
+        for line in combined_file.read_text().splitlines()
+    ]
+
+    if tmp:
+        tmp.cleanup()
+
+    return returncode == 0, out
+
+
 @__check_installed
 def version(
     check=False,
     args: List[str] = None,
-) -> Union[str, List[str]]:
+) -> Union[str, Tuple[str]]:
     """Get the rclone version number.
 
     Args:
@@ -615,30 +717,26 @@ def version(
     if check:
         args.append("--check")
 
-    process = utils.run_cmd("rclone version", args)
-
-    if process.returncode != 0:
-        raise Exception(process.stderr)
-
-    stdout = process.stdout
+    stdout, stderr = utils.run_rclone_cmd("version", args)
 
     if not check:
-        return stdout.split("\n")[0].replace("rclone ", "")
+        return stdout.splitlines()[0].replace("rclone ", "")
     else:
         yours = re.findall(r"yours:\s+([\d.]+)", stdout)[0]
-        latest = re.findall(r"latest:\s+([\d.]+)", stdout)[0]
+        latest = re.findall(r"latest:\s+([\d.]+)", stdout)
+        latest = latest[0] if latest else None
         # beta version might include dashes and word characters e.g. '1.64.0-beta.7161.9169b2b5a'
-        beta = re.findall(r"beta:\s+([.\w-]+)", stdout)[0]
+        beta = re.findall(r"beta:\s+([.\w-]+)", stdout)
+        beta = beta[0] if beta else None
+
+        if not latest or not beta:
+            logger.warning(
+                f"Failed to get latest rclone versions. The following error was output by rclone:\n{stderr}"
+            )
+
         return yours, latest, beta
 
 
-class RcloneException(ChildProcessError):
-    def __init__(self, description, error_msg):
-        self.description = description
-        self.error_msg = error_msg
-        super().__init__(f"{description}. Error message: \n{error_msg}")
-
-
 @__check_installed
 def _rclone_transfer_operation(
     in_path: str,
@@ -684,20 +782,18 @@ def _rclone_transfer_operation(
     command += utils.args2string(args)
 
     # execute the upload command
-    process = utils.rclone_progress(
+    process, errors = utils.rclone_progress(
         command,
         prog_title,
         listener=listener,
         show_progress=show_progress,
-        debug=DEBUG,
         pbar=pbar,
     )
 
     if process.wait() == 0:
-        logging.info("Cloud upload completed.")
+        logger.info("Cloud upload completed.")
     else:
-        _, err = process.communicate()
-        raise RcloneException(
+        raise utils.RcloneException(
             description=f"{command_descr} from {in_path} to {out_path} failed",
-            error_msg=err.decode("utf-8"),
+            error_msg="\n".join(errors),
         )
