@@ -1,15 +1,11 @@
 from rclone_python import rclone
 
 
-def test_about(default_test_setup):
+def test_about(tmp_local_folder):
+    # NOTE: rclone about does not work on s3 --> do locally instead
     expected_fields = set(["total", "used", "free"])
 
-    output = rclone.about(default_test_setup.remote_name)
-    assert expected_fields.issubset(set(output.keys()))
-    for field in expected_fields:
-        assert isinstance(output[field], (int, float))
-
-    output = rclone.about(default_test_setup.remote_name + ":")
+    output = rclone.about(tmp_local_folder)
     assert expected_fields.issubset(set(output.keys()))
     for field in expected_fields:
         assert isinstance(output[field], (int, float))
